{-# LANGUAGE Arrows, CPP #-}
module Controller where

import Control.Arrow
import Control.Monad (unless)
import qualified Data.IntMap as IM
import Data.Complex
import Data.IORef
import Text.Printf

import Physics
import OrbitVM
import Automata
import OSF
import Address

-- -----------------------------------------------------------

type Input = (TimeStep, SensorData)
type Controller = Automata Input DV

-- 現在時刻
time :: Automata Input Int
time = arr fst

-- センサーの値
sensor :: Automata Input SensorData
sensor = arr snd

-- スコア
score :: Automata Input Double
score = sensor >>> arr (IM.! addrScore)

-- 自衛星の位置 (地球原点)
pos :: Automata Input Pos
pos = proc i -> do
  s <- sensor -< i
  returnA -< (- s IM.! addrX) :+ (- s IM.! addrY)

-- 速度
velocity :: Automata (Pos, DV) (Maybe V)
velocity = proc (p, dv) -> do
  x <- fby Nothing -< Just (p, dv)
  returnA -<
    case x of
      Nothing -> Nothing
      Just (p', dv') ->
        let g' = mkPolar (mu / (magnitude p')**2) (phase (- p'))
            g  = mkPolar (mu / (magnitude p)**2)  (phase (- p))
            v' = p - p' - (g' + dv') / 2
            v = v' + (dv' + (g' + g) / 2)
        in Just v

velocity' :: Automata Pos (Maybe V)
velocity' = arr (\p -> (p,0)) >>> velocity

-- 自衛星の残燃料
fuelRemaining :: Automata Input Double
fuelRemaining = sensor >>> arr (IM.! addrFuelRemaining)

-- 目標とする半径
targetRadius :: Automata Input Double
targetRadius = sensor >>> arr (IM.! addrTargetRadius)

-- ターゲットの衛星の位置 (地球原点)
targetSatelitePos :: Automata Input Pos
targetSatelitePos = proc i -> do
  s <- sensor -< i
  p <- pos -< i
  let q = (s IM.! addrTargetX) :+ (s IM.! addrTargetY)
  returnA -< p + q

sim :: (Pos, V) -> Automata DV (Pos, V)
sim s0 = proc dv -> do
  rec s <- fby s0 -< simulate s dv
  returnA -< s

sim0 :: (Pos, V) -> Automata () (Pos, V)
sim0 s0 = arr (const 0) >>> sim s0

-- -----------------------------------------------------------

wait :: Int -> Task a DV ()
wait n = sequence_ $ replicate n (output 0)

hohmannTransfer :: V -> Double -> Double -> Task a DV ()
hohmannTransfer v r1 r2 = do
  output (mkPolar dv1 theta1)
  wait (t_h - 1)
  output (mkPolar dv2 theta2)
  return ()
  where
    theta1 = phase v
    theta2 = theta1 + pi
    dv1 = sqrt (mu / r1) * (sqrt (2 * r2 / (r1+r2)) - 1)
    dv2 = sqrt (mu / r2) * (1 - sqrt (2 * r1 / (r1+r2)))
    t_h = round (pi * sqrt ((r1+r2)**3 / (8*mu)))

-- -----------------------------------------------------------

control :: IOVM -> Command -> Controller -> IO [Frame]
control = controlWithTrace (\_ _ -> return ())

controlWithTrace
  :: (TimeStep -> SensorData -> IO ())
  -> IOVM -> Command -> Controller -> IO [Frame]
controlWithTrace trace vm init controller = do
  r <- newIORef newRecorder
  let step x = do
        t <- vmtime vm
        recorder <- readIORef r
        writeIORef r $! record recorder t x
        -- let x' = IM.filter (/= 0) x
        -- unless (IM.null x') $ printf "vm(t=%d): %s\n" t (show (IM.toList x'))
        vmexec vm x
  o0 <- step init
  let go c o = do
        t <- vmtime vm
        trace t o
        let score = o IM.! addrScore
        if (score == 0)
          then
            case next c (t, o) of
              (c', a) -> do
                 o' <- step (IM.fromList [(addrDVx, realPart a), (addrDVy, imagPart a)])
                 go c' o'
          else do
            recorder <- readIORef r
            putStrLn $ "Score: " ++ show score
            return $ recordFin recorder t
  go controller o0
