{-# LANGUAGE Arrows #-}
module MeetAndGreet (controller, vmtrace) where

import Control.Monad
import Control.Arrow
import Data.Maybe
import Data.Complex
import qualified Data.IntMap as IM
import System.IO
import Text.Printf

import Physics
import OrbitVM
import Automata
import Controller
import Address

import List (minimumBy, foldl)
import Data.Function (on)

import Debug.Trace

controller :: Controller
controller = proc i -> do
  t <- time -< i
  p <- pos -< i
  p2 <- targetSatelitePos -< i
  v2 <- velocity' -< p2
  rec 
    v <- velocity -< (p, dv)
    dv <- runTask_ task -< (t, p, v, p2, v2)
  returnA -< dv

task :: Task (TimeStep, Pos, Maybe V, Pos, Maybe V) DV ()
task = do
  (t, p, v, p2, v2) <- snapT
  case v of
    Nothing -> output 0 >> task
    Just v -> do
      let 
        r1 = magnitude p
        r2 = magnitude p2
        r3 = (r1 + r2) / 2
        -- ホーマン軌道の加速
        dv1 = sqrt (mu / r1) * (sqrt (2 * r2 / (r1+r2)) - 1)
        dv2 = sqrt (mu / r2) * (1 - sqrt (2 * r1 / (r1+r2)))
        -- ホーマン遷移にかかる時間
        t2 = pi * sqrt (r3**3 / mu)

        -- 角速度の符号
        s = if normalizeAngle2 (phase v - phase p) > 0 then 1 else -1
        
        -- 角速度
        w1 = s * sqrt (mu / (r1**3))
        w2 = s * sqrt (mu / (r2**3))
        -- 自分の衛星と相手の衛星が逆回転してる場合のことを考える必要はある?

        -- 現時点からのホーマン遷移の後の位相の差
        x' = (phase p2 + t2 * w2) - (phase p + s * pi)
        x = normalizeAngle (s * x')

        t1 = if abs w1 > abs w2
             then x / (abs w1 - abs w2)
             else (2*pi - x) / (abs w2 - abs w1)

      wait $ round t1
      hohmann s r2
      randevouz s

-- ホーマン
hohmann s r2 = do
  (_, p, _, _, _) <- snapT
  let r1 = magnitude p
      r3 = (r1 + r2) / 2
      dv = sqrt (mu / r1) * (sqrt (2 * r2 / (r1+r2)) - 1)
      t = pi * sqrt (r3**3 / mu)
  output $ mkPolar dv (phase p + s * (pi / 2))
  wait $ round t - 1
  orbitalize s

-- 円軌道の安定化
orbitalize s = m >> m >> m
  where
    m = do
      (_, p, Just v, _, _) <- snapT
      let theta = phase p + s * (pi / 2)
          dv2 = mkPolar (sqrt (mu / magnitude p)) theta - v
      output $ dv2

randevouz s = forever m
  where
    m = do
      (_, p, _, p2, _) <- snapT
      if ( magnitude (p2 - p) < 500 )
        then do
          constT 0 `abortWhen_` proc (_, p, _, p2, _) -> do
            returnA -< (magnitude (p2 - p) > 900)
          return ()
        else
          let r2 = magnitude p2 in
            if (s>0) == (normalizeAngle2 (phase p2  - phase p) > 0)
            then do -- 追いかける (高度を下げる)
              hohmann s (r2-800)
              return ()
            else do -- 待つ (高度を上げる)
              hohmann s (r2+800)
              return ()      

normalizeAngle :: Double -> Double
normalizeAngle a
    | a >= 2*pi = normalizeAngle (a - 2*pi)
    | a < 0     = normalizeAngle (a + 2*pi)
    | otherwise = a

normalizeAngle2 :: Double -> Double
normalizeAngle2 a
    | a > pi  = normalizeAngle2 (a - pi)
    | a < -pi = normalizeAngle2 (a + pi)
    | otherwise = a

vmtrace :: TimeStep -> SensorData -> IO ()
vmtrace t o = do
  when (t `mod` 1000 == 0) $ do
    let p1 = (- (o IM.! addrX)) :+ (- (o IM.! addrY))
        p2 = p1 + (o IM.! addrTargetX :+ o IM.! addrTargetY)
    printf "t=%d: distance=%f\n" t (magnitude (p2 - p1))
    printf "selfRadius   = %f\n" (magnitude p1)
    printf "targetRadius = %f\n" (magnitude p2)
    printf "selfPhase   = %f\n" (phase p1)
    printf "targetPhase = %f\n" (phase p2)
    hFlush stdout
    return ()

f :: Double -> Double -> Double -> Double
f r1 r2 dPhi = go r1 r2 0
  where
    val = sqrt (r2**3 * 8) / pi
    go l u n
       | n > 10 = rx
       | l>=u = rx
       | val < val' = go l rx (n+1)
       | val > val' = go rx u (n+1)
       where
         rx   = (l + u) / 2
         val' = sqrt ((r1+rx)**3) + sqrt ((r2+rx)**3)
