{-# LANGUAGE FlexibleContexts, ScopedTypeVariables, MultiParamTypeClasses #-}
module OrbitVM where

import Control.Monad (liftM, liftM2)
import Control.Monad.Trans (lift)
import Control.Monad.State (StateT (..), execStateT, modify)
import Data.Array.MArray
import qualified Data.IntMap as IM
import Data.Word
import Data.Bits
import Foreign.Marshal (alloca)
import Foreign.Storable (peek)
import System.IO

-- for specialization
import Control.Monad.ST
import Data.Array.ST
import Data.STRef
import Data.Array.IO
import Data.IORef

import Ref

type VM a r = ([Inst], a Addr Data, r Bool, r Int)
type STVM s = VM (STUArray s) (STRef s)
type IOVM = VM IOUArray IORef

type Command = IM.IntMap Data
type SensorData = IM.IntMap Data

type Data = Double
type Addr = Int -- 型変換を減らすために、Word16ではなくIntにしておく
type TimeStep = Int

data Inst
    = Add    {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Sub    {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Mult   {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Div    {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Output {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Phi    {-# UNPACK #-} !Addr {-# UNPACK #-} !Addr
    | Noop
    | Cmpz !CmpOp {-# UNPACK #-} !Addr
    | Sqrt  {-# UNPACK #-} !Addr
    | Copy  {-# UNPACK #-} !Addr
    | Input {-# UNPACK #-} !Addr
    deriving (Show)

data CmpOp
    = LTZ
    | LEZ
    | EQZ
    | GEZ
    | GTZ
    deriving (Show, Eq, Ord)

addrMax :: Int
addrMax = (1 `shiftL` 14) - 1

{-# SPECIALIZE vmexec :: STVM s -> Command -> ST s SensorData #-}
{-# SPECIALIZE vmexec :: IOVM -> Command -> IO SensorData #-}
vmexec
  :: forall m a r. (MArray a Data m, Ref r m)
  => VM a r        -- instructions, memory and status
  -> Command       -- inport
  -> m SensorData  -- outport
vmexec (inst, mem, status, time) inport = do
  i <- readRef time
  writeRef time $! i+1
  execStateT (go 0 inst) IM.empty
  where
    go :: Addr -> [Inst] -> StateT SensorData m ()
    go pc [] = seq pc $ return ()
    go pc (i:is) = seq pc $ do
      case i of
         Add  r1 r2 -> lift $ liftM2 (+) (readArray mem r1) (readArray mem r2) >>= writeArray mem pc
         Sub  r1 r2 -> lift $ liftM2 (-) (readArray mem r1) (readArray mem r2) >>= writeArray mem pc
         Mult r1 r2 -> lift $ liftM2 (*) (readArray mem r1) (readArray mem r2) >>= writeArray mem pc
         Div  r1 r2 -> lift $ liftM2 quo (readArray mem r1) (readArray mem r2) >>= writeArray mem pc
           where
             quo x y = if y==0 then 0 else x / y             
         Output r1 r2 -> do
             x <- lift $ readArray mem r2
             modify $! IM.insert (fromIntegral r1) x
         Phi r1 r2 -> lift $ do
             s <- readRef status
             readArray mem (if s then r1 else r2) >>= writeArray mem pc
         Noop -> return ()
         Cmpz op r1 -> lift $ do
             x <- readArray mem r1
             writeRef status $! evalCmpOp op x 0.0
         Sqrt  r1 -> lift $ liftM sqrt (readArray mem r1) >>= writeArray mem pc
         Copy  r1 -> lift $ readArray mem r1 >>= writeArray mem pc
         Input r1 -> lift $ writeArray mem pc $ IM.findWithDefault 0 (fromIntegral r1) inport
      go (pc+1) is
    {-# INLINE evalCmpOp #-}
    evalCmpOp :: CmpOp -> (Double -> Double -> Bool)
    evalCmpOp LTZ = (<)
    evalCmpOp LEZ = (<=)
    evalCmpOp EQZ = (==)
    evalCmpOp GEZ = (>=)
    evalCmpOp GTZ = (>)

{-# SPECIALIZE vmtime :: STVM s -> ST s TimeStep #-}
{-# SPECIALIZE vmtime :: IOVM -> IO TimeStep #-}
vmtime :: Ref r m => VM a r -> m TimeStep
vmtime (_, _, _, time) = readRef time

newMem :: MArray a Data m => [Data] -> m (a Addr Data)
newMem ds = do
  mem <- newArray (0, fromIntegral addrMax) 0
  sequence_ [ writeArray mem i d | (i, d) <- zip [0..] ds ]
  return mem

load :: FilePath -> IO ([Inst], [Data])
load fname = do
  h <- openBinaryFile fname ReadMode
  alloca $ \pdata -> alloca $ \pinst -> do
    let go evenframe is ds = seq evenframe $ do
          eof <- hIsEOF h
          if eof
            then return (reverse is, reverse ds)
            else do
              if evenframe
                then do
                  hGetBuf h pdata 8
                  hGetBuf h pinst 4
                else do
                  hGetBuf h pinst 4
                  hGetBuf h pdata 8
              d <- peek pdata
              i <- liftM decodeInst $ peek pinst
              go (not evenframe) (i : is) (d : ds)
    go True [] []

decodeInst :: Word32 -> Inst
decodeInst x =
    case d_op of
      1 -> Add  d_r1 d_r2
      2 -> Sub  d_r1 d_r2
      3 -> Mult d_r1 d_r2
      4 -> Div  d_r1 d_r2
      5 -> Output d_r1 d_r2
      6 -> Phi d_r1 d_r2
      0 ->
        case s_op of
          0 -> Noop
          1 -> Cmpz op s_r1
            where
              op = case s_imm `shiftR` 7 of
                     0 -> LTZ
                     1 -> LEZ
                     2 -> EQZ
                     3 -> GEZ
                     4 -> GTZ
                     _ -> error "unknown comparison operator"
          2 -> Sqrt s_r1
          3 -> Copy s_r1
          4 -> Input s_r1
          _ -> error "unknown S-type instruction"
      _ -> error "unknown D-type instruction"
  where
    mask14 = (1 `shiftL` 14) - 1
    mask10 = (1 `shiftL` 10) - 1
    d_op = x `shiftR` 28
    d_r1 = fromIntegral $ (x `shiftR` 14) .&. mask14
    d_r2 = fromIntegral $ x .&. mask14
    s_op = x `shiftR` 24
    s_imm = (x `shiftR` 14) .&. mask10
    s_r1  = fromIntegral $ x .&. mask14

newIOVM :: FilePath -> IO IOVM
newIOVM fname = do
  (is, ds) <- load fname
  mem <- newMem ds
  status <- newIORef False
  time <- newIORef 0
  return (is, mem, status, time)
